import asyncio
from concurrent.futures import ALL_COMPLETED, Future, wait
import logging
from threading import Thread
from typing import Any, Coroutine, Optional, Set

from aporia.consts import LOGGER_NAME


logger = logging.getLogger(LOGGER_NAME)


class EventLoop:
    """Asyncio event loop manager for worker thread loop."""

    def __init__(self):
        """Initializes an EventLoop instance."""
        logger.debug("Starting event loop in a worker thread.")
        self.loop = asyncio.new_event_loop()
        self.loop_thread = Thread(target=self.run_loop, daemon=True)

        # We keep a list of all tasks that were not awaited, to allow flushing
        # We have to do this manually to support python versions below
        # 3.7 (otherwise we could use asyncio.all_tasks())
        self.futures = []

        self.loop_thread.start()

    def run_loop(self):
        """Runs the asyncio event loop."""
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def run_coroutine(self, coro: Coroutine, await_result: Optional[bool] = True) -> Optional[Any]:
        """Runs a coroutine in the worker thread loop.

        Args:
            coro (Coroutine): Coroutine to execute
            await_result (bool, optional): If await_result is True, this will block untill the
                coroutine finishes executing. Defaults to True.

        Returns:
            Optional[Any]: Coroutine return value, if await_result is True
        """
        future = asyncio.run_coroutine_threadsafe(coro=coro, loop=self.loop)

        if await_result:
            return future.result()
        else:
            self.futures.append(future)
            return None

    def flush(self, timeout: Optional[int] = None) -> Set[Future]:
        """Waits for all currently scheduled tasks to finish.

        Args:
            timeout (int, optional): Maximum number of seconds to wait for tasks to
                complete. Default to None.

        Returns:
            Set[Future]: Set of tasks that were not completed after the timeout.
        """
        futures = self.futures
        self.futures = []

        logger.debug("Waiting for {} scheduled tasks to finish executing.".format(len(futures)))
        done, not_done = wait(futures, timeout=timeout, return_when=ALL_COMPLETED)

        logger.debug("{} tasks finished, {} tasks not finished.".format(len(done), len(not_done)))
        self.futures.extend(not_done)

        return not_done



import logging
from typing import Dict, List, Optional, Tuple, Union

from aporia.consts import LOGGER_NAME
from aporia.graphql_client import GraphQLClient


logger = logging.getLogger(LOGGER_NAME)


async def check_model_exists(
    graphql_client: GraphQLClient, model_id: str, model_version: str, timeout: int
) -> Tuple[bool, bool]:
    """Checks if a model (and version) exists.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        timeout (int): Timeout for the query

    Returns:
        Tuple[bool, bool]: (model_exists, model_version_exists)
    """
    query = """
        query CheckModelExists(
            $modelId: String!,
            $modelVersion: String!
        ) {
            modelExists(
                modelId: $modelId,
                modelVersion: $modelVersion
            ) {
                modelExists
                modelVersionExists
            }
        }
    """
    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
    }

    result = await graphql_client.query(query=query, variables=variables, timeout=timeout)
    result = result["modelExists"]
    return (result["modelExists"], result["modelVersionExists"])


async def set_features(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    feature_names: List[str],
    categorical_features: Optional[List[Union[int, str]]],
    timeout: int,
):
    """Sets model features.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        feature_names (List[str]): Feature names
        categorical_features (Optional[List[Union[int, str]]]): Categorical features
        timeout (int): Timeout for query
    """
    query = """
        mutation SetFeatures(
            $modelId: String!,
            $modelVersion: String!,
            $featureNames: [String]!,
            $categoricalFeatures: [FeatureDescriptor]
        ) {
            setFeatures(
                modelId: $modelId,
                modelVersion: $modelVersion,
                featureNames: $featureNames,
                categoricalFeatures: $categoricalFeatures
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "featureNames": feature_names,
        "categoricalFeatures": categorical_features,
    }

    result = await graphql_client.query(query=query, variables=variables, timeout=timeout)
    for warning in result["setFeatures"]["warnings"]:
        logger.warning(warning)


async def add_extra_inputs(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    extra_inputs: Dict[str, str],
    timeout: int,
):
    """Defines extra inputs for the model.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        extra_inputs (Dict[str, str]): Extra inputs to define
        timeout (int): Timeout for the query
    """
    query = """
        mutation AddExtraInputs(
            $modelId: String!,
            $modelVersion: String!,
            $extraInputs: [ExtraInput]!,
        ) {
            addExtraInputs(
                modelId: $modelId,
                modelVersion: $modelVersion,
                extraInputs: $extraInputs,
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "extraInputs": [
            {"name": input_name, "type": input_type.upper()}
            for input_name, input_type in extra_inputs.items()
        ],
    }

    result = await graphql_client.query(query=query, variables=variables, timeout=timeout)
    for warning in result["addExtraInputs"]["warnings"]:
        logger.warning(warning)


async def add_extra_outputs(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    extra_outputs: Dict[str, str],
    timeout: int,
):
    """Defines extra outputs for the model.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        extra_outputs (Dict[str, str]): Extra outputs to define
        timeout (int): Timeout for the query
    """
    query = """
            mutation AddExtraOutputs(
                $modelId: String!,
                $modelVersion: String!,
                $extraOutputs: [ExtraOutput]!,
            ) {
                addExtraOutputs(
                    modelId: $modelId,
                    modelVersion: $modelVersion,
                    extraOutputs: $extraOutputs,
                ) {
                    warnings
                }
            }
        """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "extraOutputs": [
            {"name": output_name, "type": output_type.upper()}
            for output_name, output_type in extra_outputs.items()
        ],
    }

    result = await graphql_client.query(query=query, variables=variables, timeout=timeout)
    for warning in result["addExtraOutputs"]["warnings"]:
        logger.warning(warning)



from contextlib import contextmanager
from datetime import datetime
import logging
from typing import Any, cast, Dict, Iterable, List, Optional, Union

import aporia
from aporia.api.log_actual import is_valid_log_actual_input, report_actual_batch
from aporia.api.log_json import log_json
from aporia.api.log_predict import (
    _prepare_predict_batch_input,
    generate_aporia_ids,
    is_valid_predict_input,
    log_predict_batch,
)
from aporia.api.model_setup import (
    add_extra_inputs,
    add_extra_outputs,
    check_model_exists,
    set_features,
)
from aporia.api.types import FeatureInput, FeatureValue, PredictionIdentifier
from aporia.api.utils import (
    convert_iterable_to_list_of_specified_types,
    convert_list_members_to_lists,
    convert_scalar_to_iterable,
    convert_to_list,
    convert_to_list_of_dicts,
)
from aporia.consts import LOGGER_NAME
from aporia.errors import AporiaError, handle_error
from aporia.event_loop import EventLoop
from aporia.graphql_client import GraphQLClient
from aporia.model_context import ModelContext

logger = logging.getLogger(LOGGER_NAME)


class BaseModel:
    """Base class for Model objects."""

    def __init__(
        self,
        model_id: str,
        model_version: str,
        environment: Optional[str],
        graphql_client: Optional[GraphQLClient],
        event_loop: Optional[EventLoop],
        debug: bool,
        throw_errors: bool,
        timeout: Optional[int],
    ):
        """Initializes a BaseModel object.

        Args:
            model_id (str): Model identifier
            model_version (str): Model version
            environment (str): Environment's name
            graphql_client (Optional[GraphQLClient]): GraphQL client
            event_loop (Optional[EventLoop]): AsyncIO event loop
            debug (bool): True to enable debug mode
            throw_errors (bool): True to raise exceptions in object creation and set_features
            timeout (Optional[int]): Timeout, in seconds, for synchronous functions
        """
        logger.debug(
            "Initializing model object for model {} version {}".format(model_id, model_version)
        )
        self.model_id = model_id
        self.model_version = model_version

        self._context = ModelContext()
        self._environment = cast(str, environment)
        self._graphql_client = cast(GraphQLClient, graphql_client)
        self._event_loop = cast(EventLoop, event_loop)
        self._debug = debug
        self._throw_errors = throw_errors
        self._model_ready = False
        self._model_version_exists = False
        self._timeout = cast(int, timeout)

        # Check model existence
        if event_loop is not None and graphql_client is not None:
            self._check_model_existence()

        if not self._model_ready:
            logger.error("Model object initialization failed - model operations will not work.")

    @contextmanager
    def context(
        self,
        extra_inputs: Optional[Dict[str, FeatureValue]] = None,
        extra_outputs: Optional[Dict[str, FeatureValue]] = None,
    ):
        """Sets the current model context.

        This function can be used to add specific extra_inputs/outputs to a group
        of predictions, without explicitly passing them as parameters to log_predict().

        Args:
            extra_inputs (Dict[str, FeatureValue], optional): Extra inputs, each input
                value must be a single bool, int, float or string. Defaults to None.
            extra_outputs (Dict[str, FeatureValue], optional): Extra outputs, each output
                value must be a single bool, int, float or string. Defaults to None.

        Yields:
            None

        Usage:
            The following code samples are equivalent:
            * Without context:
            >>> model.log_predict(x, y, extra_inputs={"in1": 5, "in2": 0.2}, extra_outputs={"out1": 0.01})

            * With context:
            >>> with model.context(extra_inputs={"in1": 5, "in2": 0.2}, extra_outputs={"out1": 0.01}):
            >>>     model.log_predict(x, y)
        """
        prev_context = self._context
        try:
            self._context = ModelContext(extra_inputs=extra_inputs, extra_outputs=extra_outputs)
            yield
        finally:
            self._context = prev_context

    def _check_model_existence(self):
        logger.debug(
            "Checking if model {} version {} already exists.".format(
                self.model_id, self.model_version
            )
        )
        try:
            model_exists, model_version_exists = self._event_loop.run_coroutine(
                check_model_exists(
                    graphql_client=self._graphql_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    timeout=self._timeout,
                )
            )

            if not model_exists:
                raise AporiaError("Model {} does not exist.".format(self.model_id))

            self._model_version_exists = model_version_exists
            self._model_ready = True

        except Exception as err:
            handle_error(
                message="Creating model object failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=self._throw_errors,
                original_exception=err,
            )

    def set_features(
        self,
        feature_names: Iterable[str],
        categorical: Optional[Iterable[Union[int, str]]] = None,
    ):
        """Sets feature names and categorical features for the model.

        Args:
            feature_names (Iterable[str]): List of feature names
            categorical (Iterable[Union[int, str]], optional): A list indicating which of the
                features in the 'feature_names' argument are considered categorical.
        """
        if not self._model_ready:
            return

        logger.debug("Setting model features.")
        try:
            feature_names = convert_iterable_to_list_of_specified_types(
                iterable=feature_names, name="feature_names", types=str
            )
            if categorical is not None:
                categorical = convert_iterable_to_list_of_specified_types(
                    iterable=categorical, name="categorical", types=(str, int)
                )

            self._event_loop.run_coroutine(
                set_features(
                    graphql_client=self._graphql_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    feature_names=feature_names,
                    categorical_features=categorical,
                    timeout=self._timeout,
                )
            )

            # If the query didn't fail, then the model version should now exist in the DB
            self._model_version_exists = True
        except Exception as err:
            handle_error(
                message="Setting features for {} failed, error: {}".format(self.model_id, str(err)),
                add_trace=self._debug,
                raise_exception=self._throw_errors,
                original_exception=err,
            )

    def add_extra_inputs(self, extra_inputs: Dict[str, str]):
        """Adds extra inputs to the model, which can then be reported in log_predict.

        Args:
            extra_inputs (Dict[str, str]): A mapping of input_name: input_type. type must
                be one of the following: 'numeric', 'categorical', 'boolean', 'string'.
        """
        if not self._model_ready:
            return

        logger.debug("Adding extra inputs")
        try:
            self._event_loop.run_coroutine(
                add_extra_inputs(
                    graphql_client=self._graphql_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    extra_inputs=extra_inputs,
                    timeout=self._timeout,
                )
            )
        except Exception as err:
            handle_error(
                message="Adding extra inputs failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=self._throw_errors,
                original_exception=err,
            )

    def add_extra_outputs(self, extra_outputs: Dict[str, str]):
        """Adds extra outputs to the model, which can then be reported in log_predict.

        Args:
            extra_outputs (Dict[str, str]): A mapping of output_name: output_type. type
                must be one of the following: 'numeric', 'categorical', 'boolean', 'string'.
        """
        if not self._model_ready:
            return

        logger.debug("Adding extra outputs")
        try:
            self._event_loop.run_coroutine(
                add_extra_outputs(
                    graphql_client=self._graphql_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    extra_outputs=extra_outputs,
                    timeout=self._timeout,
                )
            )
        except Exception as err:
            handle_error(
                message="Adding extra outputs failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=self._throw_errors,
                original_exception=err,
            )

    def log_predict(
        self,
        x: FeatureInput,
        y: Union[float, List[float]],
        confidence: Optional[Union[float, List[float]]] = None,
        occurred_at: Optional[datetime] = None,
        extra_inputs: Optional[Dict[str, FeatureValue]] = None,
        extra_outputs: Optional[Dict[str, FeatureValue]] = None,
        id: Optional[PredictionIdentifier] = None,
    ) -> Optional[int]:
        """Logs a single prediction.

        Args:
            x (FeatureInput): X values for all of the features in the prediction
            y (Union[float, List[float]]): Prediction result
            confidence (Union[float, List[float]], optional): Prediction confidence. Defaults
                to None.
            occurred_at (datetime, optional): Prediction timestamp. Defaults to None.
            extra_inputs (Dict[str, FeatureValue], optional): Extra inputs, each input
                value must be a single bool, int, float or string. Defaults to None.
            extra_outputs (Dict[str, FeatureValue], optional): Extra outputs, each output
                value must be a single bool, int, float or string. Defaults to None.
            id (PredictionIdentifier, optional): A unique identifier string of the prediction.
                Defaults to None.

        Returns:
            int, optional: An Aporia generated unique identifier for the prediction,
                or None if any problem occurred

        Notes:
            * The features reported in the x parameter must be set with the set_features
              method before logging predictions.
            * Extra inputs and outputs should be defined with add_extra_inputs or
              add_extra_outputsbefore they are passed to log_predict.
            * Many libraries provide a way to convert their objects to lists:
                * numpy array: Use data.tolist()
                * pandas dataframe: Use data.values.tolist()
                * scipy sparse matrix: Use data.toarray().tolist()
        """
        try:
            if confidence is not None:
                confidence = [
                    convert_to_list(convert_scalar_to_iterable(confidence))  # type: ignore
                ]

            result = self.log_predict_batch(
                x=convert_to_list_of_dicts(x),
                y=[convert_to_list(convert_scalar_to_iterable(y))],
                confidence=confidence,  # type: ignore
                occurred_at=occurred_at,
                extra_inputs=None if extra_inputs is None else [extra_inputs],
                extra_outputs=None if extra_outputs is None else [extra_outputs],
                ids=None if id is None else [id],
            )

            if isinstance(result, list) and len(result) > 0 and isinstance(result[0], int):
                return result[0]

            return None
        except Exception as err:
            handle_error(
                message="Logging prediction failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )
            return None

    def log_predict_batch(
        self,
        x: List[FeatureInput],
        y: List[List[float]],
        confidence: Optional[List[List[float]]] = None,
        occurred_at: Optional[datetime] = None,
        extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
        extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
        ids: Optional[List[PredictionIdentifier]] = None,
    ) -> Optional[List[int]]:
        """Logs multiple predictions.

        Args:
            x (List[FeatureInput]): X values for all of the features in each of the predictions
            y (List[List[float]]): Predictions results for each prediction
            confidence (List[List[float]], optional): Confidence for each prediction. Defaults to None.
            occurred_at (datetime, optional): Timestamp for each prediction. Defaults to datetime.now().
            extra_inputs (List[Dict[str, FeatureValue]], optional): Extra inputs for each prediction, each
                element in the list should a dict mapping the input name to its value, which
                can be bool, int, float or string. Defaults to None.
            extra_outputs (List[Dict[str, FeatureValue]], optional): Extra outputs for each prediction, each
                element in the list should a dict mapping the output name to its value, which
                can be bool, int, float or string. Defaults to None.
            ids (List[PredictionIdentifier], optional): An unique identifier string for each prediction.
                Defaults to None.

        Returns:
            List[int], optional: An Aporia generated unique identifier for each prediction, or None
                if any problem occurred

        Notes:
            * The features reported in the x parameter must be set with the set_features
              method before logging predictions.
            * Extra inputs and outputs should be defined with add_extra_inputs or
              add_extra_outputsbefore they are passed to log_predict_batch.
            * Many libraries provide a way to convert their objects to lists:
                * numpy array: Use data.tolist()
                * pandas dataframe: Use data.values.tolist()
                * scipy sparse matrix: Use data.toarray().tolist()
        """
        if not self._model_ready:
            return None

        if not self._model_version_exists:
            logger.warning(
                "Logging prediction failed: features for this model version were not reported."
            )
            return None

        logger.debug("Logging {} predictions".format(len(y)))
        try:
            x, y, confidence, extra_inputs, extra_outputs = _prepare_predict_batch_input(
                context=self._context,
                x=x,
                y=y,
                confidence=confidence,
                extra_inputs=extra_inputs,
                extra_outputs=extra_outputs,
            )

            if not is_valid_predict_input(
                x=x,
                y=y,
                confidence=confidence,
                extra_inputs=extra_inputs,
                extra_outputs=extra_outputs,
                ids=ids,
            ):
                logger.warning("Logging prediction failed: Invalid input.")
                return None

            aporia_ids = generate_aporia_ids(size=len(x))

            self._event_loop.run_coroutine(
                coro=self._log_predict_batch(
                    x=x,
                    y=y,
                    aporia_ids=aporia_ids,
                    confidence=confidence,
                    occurred_at=occurred_at,
                    extra_inputs=extra_inputs,
                    extra_outputs=extra_outputs,
                    ids=ids,
                ),
                await_result=False,
            )
            return aporia_ids
        except Exception as err:
            handle_error(
                message="Logging prediction batch failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )
            return None

    async def _log_predict_batch(
        self,
        x: List[FeatureInput],
        y: List[List[float]],
        aporia_ids: List[int],
        confidence: Optional[List[List[float]]] = None,
        occurred_at: Optional[datetime] = None,
        extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
        extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
        ids: Optional[List[PredictionIdentifier]] = None,
    ):
        try:
            await log_predict_batch(
                graphql_client=self._graphql_client,
                model_id=self.model_id,
                model_version=self.model_version,
                environment=self._environment,
                x=x,
                y=y,
                aporia_ids=aporia_ids,
                confidence=confidence,
                occurred_at=occurred_at,
                extra_inputs=extra_inputs,
                extra_outputs=extra_outputs,
                user_ids=ids,
            )
        except Exception as err:
            handle_error(
                message="Logging prediction failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    def log_actual(
        self,
        actuals: Union[float, List[float]],
        user_id: Optional[PredictionIdentifier] = None,
        aporia_id: Optional[int] = None,
    ):
        """Logs a single prediction actual values.

        Use this function to report the actual value (ground truth) of a previously reported prediction.
        The ground truth will be used by Aporia to calculate model's performance. Either user_id or
        aporia_id must be reported, but not both.

        Args:
            actuals (Union[float, List[float]]): Prediction actual values
            user_id (PredictionIdentifier, optional): An unique identifier of the prediction which
                was provided by the the user during the prediction report. Defaults to None.
            aporia_id (int, optional): An unique identifier of the prediction which was
                automatically generated by Aporia during the prediction report. Defaults to None.
        """
        try:
            self.log_actual_batch(
                user_ids=None if user_id is None else [user_id],
                aporia_ids=None if aporia_id is None else [aporia_id],
                actuals=[convert_to_list(convert_scalar_to_iterable(actuals))],
            )
        except Exception as err:
            handle_error(
                message="Logging actual failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    def log_actual_batch(
        self,
        actuals: List[List[float]],
        user_ids: Optional[List[PredictionIdentifier]] = None,
        aporia_ids: Optional[List[int]] = None,
    ):
        """Logs multiple predictions actual values.

            Use this function to report the actual values (ground truth) of a previously reported predictions.
            The ground truth will be used by Aporia to calculate model's performance. Either user_ids or
            aporia_ids must be reported, but not both.

        Args:
            actuals (List[List[float]]): Predictions actual values for each prediction
            user_ids (List[PredictionIdentifier], optional): A list of unique identifiers of the prediction,
                which were provided by the the user during the prediction report. Defaults to None.
            aporia_ids (List[int], optional): A list of unique identifiers of the prediction,
                which were automatically generated by Aporia during the prediction report. Defaults to None.
        """
        if not self._model_ready:
            return

        if not self._model_version_exists:
            logger.warning(
                "Logging actual failed: features for this model version were not reported."
            )
            return

        logger.debug("Logging {} actual values".format(len(actuals)))
        try:
            actuals = convert_list_members_to_lists(convert_to_list(actuals))

            if not is_valid_log_actual_input(
                user_ids=user_ids,
                aporia_ids=aporia_ids,
                actuals=actuals,
            ):
                logger.warning("Logging actual values failed: Invalid input.")
                return

            self._event_loop.run_coroutine(
                coro=self._log_actual_batch(
                    user_ids=user_ids,
                    aporia_ids=aporia_ids,
                    actuals=actuals,
                ),
                await_result=False,
            )
        except Exception as err:
            handle_error(
                message="Logging actual values batch failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    async def _log_actual_batch(
        self,
        actuals: List[List[float]],
        user_ids: Optional[List[PredictionIdentifier]] = None,
        aporia_ids: Optional[List[int]] = None,
    ):
        try:
            await report_actual_batch(
                graphql_client=self._graphql_client,
                model_id=self.model_id,
                model_version=self.model_version,
                environment=self._environment,
                user_ids=user_ids,
                aporia_ids=aporia_ids,
                actuals=actuals,
            )
        except Exception as err:
            handle_error(
                message="Logging actual batch failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    def log_json(self, data: Any):
        """Logs arbitrary data.

        Args:
            data (Any): Data to log, must be JSON serializable
        """
        if not self._model_ready:
            return

        logger.debug("Logging arbitrary data.")
        try:
            self._event_loop.run_coroutine(
                coro=self._log_json(data=data),
                await_result=False,
            )
        except Exception as err:
            handle_error(
                message="Logging arbitrary data failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )

    async def _log_json(self, data: Any):
        try:
            await log_json(
                graphql_client=self._graphql_client,
                model_id=self.model_id,
                model_version=self.model_version,
                environment=self._environment,
                data=data,
            )
        except Exception as err:
            handle_error(
                message="Logging arbitrary data failed, error: {}".format(str(err)),
                add_trace=self._debug,
                raise_exception=False,
                original_exception=err,
                log_level=logging.ERROR,
            )


class Model(BaseModel):
    """Model object."""

    def __init__(self, model_id: str, model_version: str):
        """Initializes a Model object.

        Args:
            model_id (str): Model identifier, as received from the Aporia dashboard.
            model_version (str): Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
        """
        if aporia.context is None:
            logger.error("Aporia was not initialized.")
            super().__init__(
                model_id=model_id,
                model_version=model_version,
                graphql_client=None,
                environment=None,
                event_loop=None,
                debug=False,
                throw_errors=False,
                timeout=None,
            )
        else:
            super().__init__(
                model_id=model_id,
                model_version=model_version,
                graphql_client=aporia.context.graphql_client,
                event_loop=aporia.context.event_loop,
                environment=aporia.context.config.environment,
                debug=aporia.context.config.debug,
                throw_errors=aporia.context.config.throw_errors,
                timeout=aporia.context.config.blocking_call_timeout,
            )



from typing import List, Optional, Union
from aporia.api.types import FeatureValue, PredictionIdentifier, FeatureInput


def is_valid_predict_param_features(data: List[FeatureInput]) -> bool:
    """Verify input features data parameter is in valid input format.

    Args:
        data (List[FeatureInput]): Features data parameter

    Returns:
        bool: True if the features data parameter is in valid input format, False otherwise
    """
    if not isinstance(data, list):
        return False

    if len(data) == 0:
        return False

    if not all((isinstance(data_point, dict) and len(data_point) > 0) for data_point in data):
        return False

    return True


def is_valid_predict_param_list(data: Union[List[List[float]], List[List[FeatureValue]]]) -> bool:
    """Verify input data list is in valid input format.

    Args:
        data (Union[List[List[float]], List[List[FeatureValue]]]): Data list

    Returns:
        bool: True if the data list is in valid input format, False otherwise
    """
    if not isinstance(data, list):
        return False

    if len(data) == 0:
        return False

    if not all((isinstance(data_point, list) and len(data_point) > 0) for data_point in data):
        return False

    return True


def is_valid_ids(
    data: Optional[Union[List[PredictionIdentifier], List[int]]] = None,
    expected_length: Optional[int] = None,
) -> bool:
    """Verify input identifiers list is in valid input format.

    Args:
        data (Union[List[PredictionIdentifier], List[int]], optional): Input list of
            identifiers. Defaults to None.
        expected_length (int, optional): The expected length of the list. Defaults to None.

    Returns:
        bool: True if the identifiers list is in valid input format, False otherwise
    """
    if data is None:
        return False

    if not isinstance(data, list):
        return False

    if expected_length is not None and len(data) != expected_length:
        return False

    if not all((isinstance(data_point, int) or isinstance(data_point, str)) for data_point in data):
        return False

    return True



from datetime import datetime
import logging
from typing import Dict, List, Optional

import simpleflake
import ujson

from aporia.api.input_validation import (
    is_valid_ids,
    is_valid_predict_param_features,
    is_valid_predict_param_list,
)
from aporia.api.types import FeatureInput, FeatureValue, PredictionIdentifier
from aporia.api.utils import (
    convert_list_members_to_lists,
    convert_to_list,
    convert_to_list_of_dicts,
)
from aporia.consts import LOGGER_NAME
from aporia.graphql_client import GraphQLClient
from aporia.model_context import ModelContext

logger = logging.getLogger(LOGGER_NAME)


async def log_predict_batch(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    environment: str,
    x: List[FeatureInput],
    y: List[List[float]],
    aporia_ids: List[int],
    confidence: Optional[List[List[float]]] = None,
    occurred_at: Optional[datetime] = None,
    extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
    extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
    user_ids: Optional[List[PredictionIdentifier]] = None,
):
    """Reports a batch of predictions.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        environment (str): Environment in which aporia is running.
        x (List[FeatureInput]): Feature values
        y (List[List[float]]): Prediction result
        aporia_ids (List[int]): List of unique identifiers generated by Aporia for each prediction
        confidence (List[List[float]], optional): Prediction confidence. Defaults to None.
        occurred_at (datetime, optional): Prediction timestamp. Defaults to None.
        extra_inputs (List[Dict[str, FeatureValue]], optional): Extra inputs. Defaults to None.
        extra_outputs (List[Dict[str, FeatureValue]], optional): Extra outputs. Defaults to None.
        user_ids (List[PredictionIdentifier], optional): List of unique identifiers provided by the user for
            each prediction. Defaults to None.
    """
    query = """
        mutation LogPredict(
            $modelId: String!,
            $modelVersion: String!,
            $x: JSONString!,
            $yPred: [[Float]]!,
            $confidence: [[Float]],
            $occurredAt: String,
            $environment: String!,
            $aporiaIds: [BigInt]!,
            $extraInputs: [[ExtraInputValue]],
            $extraOutputs: [[ExtraOutputValue]],
            $userIds: [PredictionIdentifier]
        ) {
            logPredictions(
                modelId: $modelId,
                modelVersion: $modelVersion,
                x: $x,
                yPred: $yPred,
                confidence: $confidence,
                occurredAt: $occurredAt,
                environment: $environment,
                aporiaIds: $aporiaIds,
                extraInputs: $extraInputs,
                extraOutputs: $extraOutputs,
                userIds: $userIds
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "x": ujson.dumps(x),
        "yPred": y,
        "confidence": confidence,
        "occurredAt": None if occurred_at is None else occurred_at.isoformat(),
        "environment": environment,
        "aporiaIds": aporia_ids,
        "extraInputs": _build_extra_io_values(extra_inputs),
        "extraOutputs": _build_extra_io_values(extra_outputs),
        "userIds": user_ids,
    }

    result = await graphql_client.query_with_retries(query, variables)
    for warning in result["logPredictions"]["warnings"]:
        logger.warning(warning)


def _prepare_predict_batch_input(
    context: ModelContext,
    x: List[FeatureInput],
    y: List[List[float]],
    confidence: Optional[List[List[float]]] = None,
    extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
    extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
) -> tuple:
    x = convert_to_list_of_dicts(x)
    y = convert_list_members_to_lists(convert_to_list(y))
    if confidence is not None:
        confidence = convert_list_members_to_lists(convert_to_list(confidence))

    extra_inputs = _prepare_batch_input_extras(extra_inputs, context.extra_inputs, len(y))
    extra_outputs = _prepare_batch_input_extras(extra_outputs, context.extra_outputs, len(y))

    return x, y, confidence, extra_inputs, extra_outputs


def _prepare_batch_input_extras(
    extras: Optional[List[Dict[str, FeatureValue]]],
    context_extras: Optional[Dict[str, FeatureValue]],
    expected_length: int,
) -> Optional[List[Dict[str, FeatureValue]]]:
    if context_extras is not None:
        if extras is not None:
            # We want the extra_input/output parameters to override the values in the
            # context, so they must be second in the dict merge.
            return [{**context_extras, **extra} for extra in extras]
        else:
            return [context_extras for _ in range(expected_length)]

    return extras


def _build_extra_io_values(
    data: Optional[List[Dict[str, FeatureValue]]] = None
) -> Optional[List[List[dict]]]:
    if data is None:
        return None

    result = []
    for data_point in data:
        result.append([{"name": name, "value": value} for name, value in data_point.items()])

    return result


def is_valid_predict_input(
    x: List[FeatureInput],
    y: List[List[float]],
    confidence: Optional[List[List[float]]] = None,
    extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
    extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
    ids: Optional[List[PredictionIdentifier]] = None,
) -> bool:
    """Checks if log_predict_batch input is valid.

    Args:
        x (List[FeatureInput]): Feature values
        y (List[List[float]]): Prediction result
        confidence (List[List[float]], optional): Prediction confidence. Defaults to None.
        extra_inputs (List[Dict[str, FeatureValue]], optional): Extra inputs. Defaults to None.
        extra_outputs (List[Dict[str, FeatureValue]], optional): Extra outputs. Defaults to None.
        ids (List[PredictionIdentifier], optional): An unique identifier string of the
            prediction. Defaults to None.

    Returns:
        bool: True if all of the parameters are valid. False otherwise
    """
    if not is_valid_predict_param_features(x):
        logger.warning(
            "Invalid input format for x parameter - expected a non-empty list of dictionaries"
        )
        return False

    if not is_valid_predict_param_list(y):
        logger.warning(
            "Invalid input format for y parameter - expected a non-empty 2 dimensional list"
        )
        return False

    if len(x) != len(y):
        logger.warning("Invalid input: x and y should have identical length")
        return False

    if confidence is not None:
        if not is_valid_predict_param_list(confidence):
            logger.warning(
                "Invalid input format for confidence parameter - expected a "
                "non-empty 2 dimensional list"
            )
            return False

        if len(y) != len(confidence):
            logger.warning("Invalid input: y and confidence should have identical length")
            return False

    if extra_inputs is not None:
        if not _is_valid_extra_io(data=extra_inputs, expected_length=len(x)):
            logger.warning(
                "Invalid input: extra_inputs must be a list of values equal in length to the "
                "number of predictions, such that each element in the list is a dict "
                "containing the extra inputs for a single prediction"
            )
            return False

    if extra_outputs is not None:
        if not _is_valid_extra_io(data=extra_outputs, expected_length=len(x)):
            logger.warning(
                "Invalid input: extra_outputs must be a list of values equal in length to the "
                "number of predictions, such that each element in the list is a dict "
                "containing the extra outputs for a single prediction"
            )
            return False

    if ids is not None:
        if not is_valid_ids(data=ids, expected_length=len(x)):
            logger.warning(
                "Invalid input: ids must be a list of strings or ints equal "
                "in length to the number of predictions"
            )
            return False

    return True


def _is_valid_extra_io(data: List[Dict[str, FeatureValue]], expected_length: int) -> bool:
    if not isinstance(data, list):
        return False

    if len(data) != expected_length:
        return False

    if not all(isinstance(data_point, dict) for data_point in data):
        return False

    return True


def generate_aporia_ids(size: int) -> List[int]:
    """Generates a list of unique identifiers.

    Args:
        size (int): Size of the generated list

    Returns:
        List[int]: A list of unique identifiers in the specified size
    """
    ids = set()  # type: ignore
    # We want to make sure there are no duplications
    while len(ids) < size:
        ids.add(simpleflake.simpleflake())
    return list(ids)



"""Aporia SDK.

Usage:
    >>> import aporia
    >>> aporia.init(
    >>>     token='12345',
    >>>     host='production.myorg.cloud.aporia.com',
    >>>     port=443,
    >>>     environment="production"
    >>> )
    >>> model = aporia.Model(model_name='my-model'], model_version='v1')
    >>> model.set_features(feature_names=['x1', 'x2', 'x3'], categorical=['x2'])
    >>> model.log_predict(x={'x1': 1.3, 'x2': 0.7, 'x3': 0.2}, y=[0.09])
"""
import atexit
from contextlib import suppress
import logging
import sys
from typing import Optional

from aporia.config import Config
from aporia.consts import LOG_FORMAT, LOGGER_NAME
from aporia.errors import handle_error
from aporia.event_loop import EventLoop
from aporia.graphql_client import GraphQLClient
from aporia.model import Model

try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"


__all__ = ["init", "flush", "Model"]


class Context:
    """Global context."""

    def __init__(self, graphql_client: GraphQLClient, event_loop: EventLoop, config: Config):
        """Initializes the context.

        Args:
            graphql_client (GraphQLClient): GraphQL client.
            event_loop (EventLoop): Event loop.
            config (Config): Aporia config.
        """
        self.graphql_client = graphql_client
        self.event_loop = event_loop
        self.config = config


context = None
logger = logging.getLogger(LOGGER_NAME)


def init(
    token: Optional[str] = None,
    host: Optional[str] = None,
    environment: Optional[str] = None,
    port: Optional[int] = None,
    debug: Optional[bool] = None,
    throw_errors: Optional[bool] = None,
    blocking_call_timeout: Optional[int] = None,
):
    """Initializes the Aporia SDK.

    Args:
        token (str): Authentication token.
        host (str): Controller host.
        environment (str): Environment in which aporia is initialized (e.g production, staging).
        port (int, optional): Controller port. Defaults to 443.
        debug (bool, optional): True to enable debug mode - this will cause additional logs
            and stack traces during exceptions. Defaults to False.
        throw_errors (bool, optional): True to cause errors to be raised
            as exceptions. Defaults to False.
        blocking_call_timeout (int, optional): Timeout, in seconds, for blocking aporia API
            calls - Model(), set_features(), add_extra_inputs(), add_extra_outputs().

    Notes:
        * The token, host and environment parameters are required.
        * All of the parameters here can also be defined as environment variables:
            * token -> APORIA_TOKEN
            * host -> APORIA_HOST
            * environment -> APORIA_ENVIRONMENT
            * port -> APORIA_PORT
            * debug -> APORIA_DEBUG
            * throw_errors -> APORIA_THROW_ERRORS
            * blocking_call_timeout -> APORIA_BLOCKING_CALL_TIMEOUT
        * Values passed as parameters to aporia.init() override the values from
          the corresponding environment variables.
    """
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if not logger.hasHandlers():
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT, style="{"))
        logger.addHandler(handler)

    logger.debug("Initializing Aporia SDK.")

    try:
        config = Config(
            token=token,
            host=host,
            environment=environment,
            port=port,
            debug=debug,
            throw_errors=throw_errors,
            blocking_call_timeout=blocking_call_timeout,
        )

        # Init graphql client and event loop
        event_loop = EventLoop()
        graphql_client = GraphQLClient(token=config.token, host=config.host, port=config.port)
        event_loop.run_coroutine(graphql_client.open())

        global context
        context = Context(graphql_client=graphql_client, event_loop=event_loop, config=config)

        atexit.register(shutdown)
        logger.debug("Aporia SDK initialized.")
    except Exception as err:
        handle_error(
            message="Initializing Aporia SDK failed, error: {}".format(str(err)),
            add_trace=False if debug is None else debug,
            raise_exception=False if throw_errors is None else throw_errors,
            original_exception=err,
        )


def flush(timeout: Optional[int] = None) -> int:
    """Waits for all of the prediction logs to reach the controller.

    Args:
        timeout (int, optional): Maximum number of seconds to wait for the flush to
            complete. Defaults to None (No time limit).

    Returns:
        int: Number of tasks left in the queue after the flush.

    Notes:
        * If the timeout is exceeded, there may be some unfinished tasks left in the queue.
          To make sure that everything is finished, call flush() without timeout.
        * The return value does not include tasks that were added during or
          after the call to flush().
    """
    if context is None:
        logger.error("Flush failed, Aporia SDK was not initialized.")
        return 0

    logger.debug("Flushing remaining data")
    not_done = context.event_loop.flush(timeout=timeout)
    return len(not_done)


def shutdown():
    """Shuts down the Aporia SDK.

    Notes:
        * It is advised to call flush() before calling shutdown(), to ensure that
          all of the data that was sent reaches the controller.
    """
    with suppress(Exception):
        global context
        if context is not None:
            context.event_loop.run_coroutine(context.graphql_client.close())
            context = None



import collections.abc
from typing import Any, Iterable, List, Tuple, Union


def convert_scalar_to_iterable(data: Any) -> Iterable[Any]:
    """Converts a scalar to an iterable.

    Args:
        data (Any): Python or numpy scalar

    Raises:
        RuntimeError: If data cannot be converted

    Returns:
        Iterable[Any]: An iterable containing the input scalar
    """
    # Input is already iterable, do nothing
    if isinstance(data, collections.abc.Iterable):
        return data
    # Python scalar (using type instead of isinstance, because bool is a subclass of int)
    elif type(data) in (int, float):
        return [data]
    # Numpy scalar
    elif hasattr(data, "tolist"):
        return [data.tolist()]

    raise RuntimeError("Unsupported input scalar type {}".format(type(data)))


def convert_to_list(data: Any) -> List[Any]:
    """Converts an object to a list.

    Args:
        data (Any): A list, numpy array or pandas DataFrame

    Raises:
        RuntimeError: If data cannot be converted

    Returns:
        List[Any]: The original data, converted to a python list
    """
    # Data is already a list, do nothing
    if isinstance(data, list):
        return data
    # Numpy ndarray
    elif hasattr(data, "tolist"):
        return data.tolist()
    # Pandas DataFrame, Series, Index
    elif hasattr(data, "values") and hasattr(data.values, "tolist"):
        return data.values.tolist()

    raise RuntimeError("Unsupported input type {}".format(type(data)))


def convert_list_members_to_lists(data: List[Any]) -> List[List[Any]]:
    """Converts the members of a list to lists.

    Args:
        data (List[Any]): List whose members should be converted

    Returns:
        List[List[Any]]: A list containing the converted members of data.
    """
    # To avoid a high computational overhead, we assume all list members are of the same type
    if not isinstance(data[0], list):
        return [convert_to_list(member) for member in data]

    return data


def convert_iterable_to_list_of_specified_types(
    iterable: Iterable[Any], name: str, types: Union[type, Tuple[type, ...]]
) -> List[Any]:
    """Converts an iterable to a list, with type validation.

    Args:
        iterable (Iterable[Any]): Iterable to convert.
        name (str): Name of the iterable, for error messages.
        types (Union[type, Tuple[type]]): Expected types for elements in the iterable.

    Raises:
        RuntimeError: If the input is not an iterable.
        RuntimeError: If the elements of the iterable are not of the expected types.

    Returns:
        List[Any]: A list containing the elements from the iterable.
    """
    if not isinstance(iterable, collections.abc.Iterable):
        raise RuntimeError("Invalid input: {} must be iterable".format(name))

    if not all(isinstance(member, types) for member in iterable):
        raise RuntimeError(
            "Invalid input: {} must only contain elements of types {}".format(name, types)
        )

    return list(iterable)


def convert_to_list_of_dicts(data: Any) -> List[dict]:
    """Converts an object to a list of dicts.

    Args:
        data (Any): A list, dict or pandas DataFrame

    Raises:
        RuntimeError: If data cannot be converted

    Returns:
        List[Any]: The original data, converted to a python list of dicts
    """
    if isinstance(data, list):
        return data
    elif isinstance(data, dict):
        return [data]
    # Pandas DataFrame
    elif hasattr(data, "to_dict"):
        return data.to_dict("records")

    raise RuntimeError("Unsupported input type {}".format(type(data)))



import logging
from typing import List, Optional

from aporia.api.input_validation import is_valid_ids, is_valid_predict_param_list
from aporia.api.types import PredictionIdentifier
from aporia.consts import LOGGER_NAME
from aporia.graphql_client import GraphQLClient

logger = logging.getLogger(LOGGER_NAME)


async def report_actual_batch(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    environment: str,
    actuals: List[List[float]],
    user_ids: Optional[List[PredictionIdentifier]] = None,
    aporia_ids: Optional[List[int]] = None,
):
    """Reports a batch of predictions.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version
        environment (str): Environment in which aporia is running.
        actuals (List[List[float]]): Predictions actual values for each prediction
        user_ids (List[PredictionIdentifier], optional): A list of unique identifiers of the prediction,
            which were provided by the the user during the prediction report. Defaults to None.
        aporia_ids (List[int], optional): A list of unique identifiers of the prediction,
            which were automatically generated by Aporia during the prediction report. Defaults to None.
    """
    query = """
        mutation LogActual(
            $modelId: String!,
            $modelVersion: String!,
            $environment: String!,
            $userIds: [PredictionIdentifier],
            $aporiaIds: [BigInt],
            $actuals: [[Float]]!
        ) {
            logActual(
                modelId: $modelId,
                modelVersion: $modelVersion,
                environment: $environment,
                userIds: $userIds,
                aporiaIds: $aporiaIds,
                actuals: $actuals
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "actuals": actuals,
        "environment": environment,
        "userIds": user_ids,
        "aporiaIds": aporia_ids,
    }

    result = await graphql_client.query_with_retries(query, variables)
    for warning in result["logActual"]["warnings"]:
        logger.warning(warning)


def is_valid_log_actual_input(
    actuals: List[List[float]],
    user_ids: Optional[List[PredictionIdentifier]] = None,
    aporia_ids: Optional[List[int]] = None,
) -> bool:
    """Checks if log_actual_batch input is valid.

    Args:
        actuals (List[List[float]]): Predictions actual values for each prediction
        user_ids (List[PredictionIdentifier], optional): A list of unique identifiers of the prediction,
            which were provided by the the user during the prediction report. Defaults to None.
        aporia_ids (List[int], optional): A list of unique identifiers of the prediction,
            which were automatically generated by Aporia during the prediction report. Defaults to None.

    Returns:
        bool: True if all of the parameters are valid. False otherwise
    """
    if not is_valid_predict_param_list(actuals):
        logger.debug("Invalid input format for actual parameter")
        return False

    if (user_ids is None or len(user_ids) == 0) and (aporia_ids is None or len(aporia_ids) == 0):
        logger.debug("Invalid input: neither user ids nor aporia ids were provided")
        return False

    user_ids_len = 0 if user_ids is None else len(user_ids)
    aporia_ids_len = 0 if aporia_ids is None else len(aporia_ids)

    if user_ids_len > 0 and aporia_ids_len > 0:
        logger.debug("Invalid input: user ids and aporia ids cannot be used together")
        return False

    if is_valid_ids(user_ids, len(actuals)):
        return True
    elif is_valid_ids(aporia_ids, len(actuals)):
        return True
    logger.debug("Invalid input: ids must be a list equal in length to the number of predictions")
    return False



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aporia', 'aporia.api']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0',
 'simpleflake>=0.1.5,<0.2.0',
 'tenacity>=6.2.0,<7.0.0',
 'ujson==3.2.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.5.0,<2.0.0']}

setup_kwargs = {
    'name': 'aporia',
    'version': '1.0.21',
    'description': 'Aporia SDK',
    'long_description': '# Aporia SDK',
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aporia-ai/sdk',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.5.3,<4.0.0',
}


setup(**setup_kwargs)



from typing import Dict, Optional

from aporia.api.types import FeatureValue


class ModelContext:
    """Context object for prediction syntax sugar."""

    __slots__ = ["extra_inputs", "extra_outputs"]

    def __init__(
        self,
        extra_inputs: Optional[Dict[str, FeatureValue]] = None,
        extra_outputs: Optional[Dict[str, FeatureValue]] = None,
    ):
        """Initializes a ModelContext object.

        Args:
            extra_inputs (Dict[str, FeatureValue], optional): Extra inputs. Defaults to None.
            extra_outputs (Dict[str, FeatureValue], optional): Extra outputs. Defaults to None.
        """
        self.extra_inputs = extra_inputs
        self.extra_outputs = extra_outputs



import logging

import aiohttp
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

from aporia.consts import LOGGER_NAME
from aporia.errors import AporiaError

logger = logging.getLogger(LOGGER_NAME)


QUERY_MAX_ATTEMPTS = 4
QUERY_TIMEOUT = 10
QUERY_RETRY_INITIAL_SLEEP = 2
DEFAULT_TIMEOUT_SEC = 300


class GraphQLClient:
    """Asynchronous graphql client."""

    def __init__(self, token: str, host: str, port: int):
        """Initialize a GraphQLClient instance.

        Args:
            token (str): Authorization token
            host (str): Controller address
            port (int): Controller port
        """
        logger.debug("Initializing GraphQL client.")
        if not host.startswith("http"):
            host = "https://{}".format(host)

        self.request_url = "{}:{}/v1/controller/graphql".format(host, port)
        self.headers = {"Authorization": "Bearer {}".format(token)}
        self.session = None

    async def open(self):
        """Opens the http session.

        Notes:
            * This must be executed in the event loop
        """
        logger.debug("Creating HTTP session with controller.")
        self.session = aiohttp.ClientSession(headers=self.headers)

    @retry(
        stop=stop_after_attempt(QUERY_MAX_ATTEMPTS),
        wait=wait_exponential(min=QUERY_RETRY_INITIAL_SLEEP),
        retry=retry_if_exception(lambda err: not isinstance(err, AporiaError)),
        reraise=True,
    )
    async def query_with_retries(
        self, query: str, variables: dict, timeout: int = DEFAULT_TIMEOUT_SEC
    ) -> dict:
        """Executes a GraphQL query with retries in case of failure.

        Args:
            query (str): GraphQL query string
            variables (dict): Variables for the query
            timeout (int): Timeout for the entire request, in seconds. Defaults to 5 minutes.

        Returns:
            dict: GraphQL query result
        """
        return await self.query(query, variables, timeout)

    async def query(self, query: str, variables: dict, timeout: int = DEFAULT_TIMEOUT_SEC) -> dict:
        """Executes a GraphQL query and returns the result.

        Args:
            query (str): GraphQL query string
            variables (dict): Variables for the query
            timeout (int): Timeout for the entire request, in seconds. Defaults to 5 minutes.

        Returns:
            dict: GraphQL query result
        """
        logger.debug("Sending GraphQL query: {}, variables: {}".format(query, variables))
        # Note: By default, aiohttp uses SSL and verifies the certificate in HTTPS requests
        async with self.session.post(  # type: ignore
            url=self.request_url,
            json={"query": query, "variables": variables},
            timeout=aiohttp.ClientTimeout(total=timeout),
        ) as response:

            if response.status != 200:
                if response.status == 400:
                    errors = (await response.json())["errors"]  # type: ignore
                    raise AporiaError("Server Error: {}".format(errors[0]["message"]))

                elif response.status == 401:
                    raise AporiaError("Authentication failed, please check your token.")

                else:
                    raise AporiaError("Unexpected HTTP error {}".format(response.status))

            return (await response.json())["data"]  # type: ignore

    async def close(self):
        """Closes the http session."""
        await self.session.close()
